UNIT HTimers;

(**********************************************************************
  High Resolution Timer Unit for Turbo-Pascal 6.0 and 7.0
  (c) copyright 1992-97 by Klaus Hartnegg, hartnegg@iname.com

  Features:
    This clock ticks 1193181 times per second and is completely
    independend of the CPU frequency.

  Limit:
    DO NOT USE THIS IN ANY VERSION OF WINDOWS OR OS/2
    OR WITH ANY MULTITASKING ENVIRONMENT

  Resolution:
    1/1000 (milli seconds, tousands of a second) of times of up 180 seconds
    (last digit will always be zero after more than 180 seconds).
    1/100 second of times up to about 30 minutes (overrun afterwards).
    Very short times can be even measured with a resolution of
    1/1000000 (a millionth) of a second but this should be used with
    care (runtime error if the time is too long).

  Method:
    This unit reads the Bios Time Couter (a longint variable stored
    at memory $40:$6C) and the value of channel 0 of the PIT
    (PIT=programmable interrupt counter). The PIT is reprogrammed
    to mode 2.

  Technical Explanation:
    PIT channel 0 counts from 65535 backwards to 0 with 1193181 ticks
    per second. Each time it restarts the Bios Counter is incremented
    (18.2 times per second). If you read only the DOS time, you will
    get a value calculated from the Bios Time Counter, i.e. a value
    with a resolution of 18.2 ticks per second. The resolution can be
    improved by a factor of 65536 by taking the current value of the
    PIT counter 0 into account.

    This unit reprograms PIT counter 0 to mode 2. Reason:
    The PIT counter 0 can be operated in two modes: 2 or 3. In mode 3
    it counts twice in steps of two and the Bios Timer Counter is
    increased only every second time. This was done to increase the
    signal duration that triggers the interrupt that increases the
    Bios Time Counter but I have never seen any PC that really requires
    this, not even the oldest ones. Unfortunately old versions of the
    PIT do not allow to determine if they are in the first or the
    second countdown while running in mode 3. Reprogramming to mode
    2 doesn't cause any problems on any modern PC. In fact many
    new PC's have the PIT already running in mode 2.


  Installation:
    Simply include the name of this unit with 'uses'. Everything
    else is done automatically:
    The initialization part of the unit calls Init.
    Init reprograms the PIT to mode 2.
    The exit procedure calls done. Done restores the old PIT mode.

  Example:

    program test;
    uses htimers;
    var
      timer : htimerobj;
      time  : longint;
    begin
      timer.setzero;
      { do anything here }
      time := timer.getmsecs;

      writeln ('This took ',time,' milliseconds');
    end.

  If you want more precise timing, take the time for
  the time measurement into account and subtract it:

    program test;
    uses htimers;
    var
      timer : htimerobj;
      time,
      time0  : longint;
    begin
      timer.setzero;
      time0 := timer.getmsecs;

      timer.setzero;
      { do anything here }
      time := timer.getmsecs;

      writeln ('This took ',time-time0,' milliseconds');
    end.

**********************************************************************)



INTERFACE

uses
  PIT;

var
  BiosTimeCounter  : longint absolute $0040:$006C; { # int $1C, Dos Clock  }


Procedure Install;
Procedure Restore;
Function  IsInstalled : boolean;

type
  HTimerObj = object
            {------------------------------}
             HiOffset, LoOffset : longint;
            {------------------------------}
             Procedure SetZero;
             Function  GetTicks : LongInt;
             Function  GetMSecs : LongInt;
             Function  GetMuSecs: LongInt;
             { last digit always 0 after more than 180 seconds }
            {------------------------------}
           end;



IMPLEMENTATION

{ Calling install many times will in the mean not changed the dos clock.
  But restoring mode 3 delay the dos clock by one tick half of the times.
}

const
  _IsInstalled : boolean = false;
  OldExit      : pointer = Nil;
  OldPitMode   : byte = 3;


Procedure MyExit; FAR;
begin
  Exitproc := OldExit;
  Restore;
end;


Procedure Install;
begin
   if _IsInstalled then exit;
   _IsInstalled := true;

   if GetPitType = PIT_8254
      then OldPitMode := GetPit0Mode
      else OldPitMode := 3;

   SetPit0Mode (2,0);
   if OldExit = Nil then begin
      OldExit  := ExitProc;
      ExitProc := @MyExit;
   end;
end;


Procedure Restore;
begin
   if not _IsInstalled then exit;
   _IsInstalled := false;
   SetPit0Mode (OldPitMode,0);
   if Exitproc = @MyExit then begin
      ExitProc := OldExit;
      OldExit  := Nil;
   end;
end;


Function IsInstalled : boolean;
begin
  IsInstalled := _IsInstalled;
end;


Procedure GetTimers (var dos_val, pit_val : longint);
{ Read both timers and make sure that not one of them has just }
{ been increased during readout but the other not yet }
var i1,i2,j1,j2 : longint;
begin
  i1 := BiosTimeCounter;
  j1 := GetPit0Count;
  i2 := BiosTimeCounter;
  j2 := GetPit0Count;

  if i1 = i2 then begin  { first readout was ok }
     dos_val := i1;
     pit_val := j1;
  end
  else begin             { clock has ticked meanwhile }
     dos_val := i2;      { then the second values should be ok }
     pit_val := j2;
  end;
end;



Procedure HTimerObj.SetZero;
begin
  GetTimers (HiOffset, LoOffset);
end;



Function HTimerObj.GetTicks : LongInt;
var DosTicks, HiresTicks : longint;
begin
  GetTimers (DosTicks, HiresTicks);
  GetTicks := (DosTicks - HiOffset) shl 16 + (LoOffset - HiresTicks);
end;



Function HTimerObj.GetMSecs : longint;
const
  exp10 : array [1..6] of longint = (1,10,100,1000,10000,100000);
var
  ticks : longint;
  help  : longint;
  step  : byte;
begin
  Ticks := GetTicks;

  help  := HTicks_per_second;
  step  := 6;
  while ticks > MaxLongInt div 1000 do begin
     ticks := ticks div 10;
     help  := help  div 10;
     if step > 1 then dec (step)
  end;
  ticks := (1000 * Ticks) div help;

{ if ticks > 180000 then ticks := 10 * (ticks div 10); }
  if ticks > exp10[step] then ticks := exp10[step] * (ticks div exp10[step]);

  GetMSecs := ticks;
end;


Function HTimerObj.GetMuSecs : LongInt;
begin
  GetMuSecs := round( (1000 * 1000 / HTicks_per_second) * GetTicks);
end;



BEGIN
  Install;
END.
